-------------------------------------------------------------------------
-- DeTrashLoot - Delete or sell items in bags based on their market value
-------------------------------------------------------------------------
-- Modules/Item.lua - Item Stuff
--
-- Author: Expelliarm5s / June 2022 / All Rights Reserved
--
-- Version 1.1.7
-------------------------------------------------------------------------
-- luacheck: ignore 212 globals DLAPI
-- luacheck: globals GetDetailedItemLevelInfo OEMarketInfo

local addonName, addon = ...
local Item = addon:NewModule("Item", "AceConsole-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(addonName)
local private = {}
-------------------------------------------------------------------------------------------------------

------------------------------------------------------------------------------
-- Debug Stuff

function Item:DebugPrintf(...)
	if addon.isDebug then
		local status, res = pcall(format, ...)
		if status then
			addon:DebugLog("Itm~" .. res)
		end
	end
end

------------------------------------------------------------------------------
-- Login

function Item:Login()
	Item:DebugPrintf("Login()")
end

-----------------------------------------------------------------------------
-- Helper Functions

-- Return an escaped link for diagnose
function Item:EscapedLink(link)
	local l = ""
	if type(link) == "string" and link:match("^\124(.+)\124h") then
		l = gsub(link, "\124", "\124\124")
	end
	return l
end

-- Return itemID from itemLink
function Item:GetItemID(val)
	if type(val) == "number" then
		return val
	end

	if type(val) == "string" then
		local id = val:match("Hitem:([0-9]+)")
		if id and tonumber(id) then
			return tonumber(id)
		end
	end
end

-- Return itemName from itemLink
function Item:GetItemName(val)
	if type(val) == "string" then
		local id = val:match("|h(.+)|h")
		if id then
			return id
		end
	end
end

-- Return money string from copper
function Item:FromCopper(copper, isVS)
	if not tonumber(copper) then return end

	local g = floor(copper / COPPER_PER_GOLD)
	local s = floor((copper - (g * COPPER_PER_GOLD)) /  COPPER_PER_SILVER)
	local c = copper - (g * COPPER_PER_GOLD) - (s * COPPER_PER_SILVER)

	local result = ""

	local vsC = ""
	local vsR = ""
	if isVS then
		vsC = ITEM_QUALITY_COLORS[7].hex
		vsR = FONT_COLOR_CODE_CLOSE
	end

	if not addon.db.global.discardSilver then
		if not addon.db.global.discardCopper then
			if (s > 0) or (g > 0) then
				result = vsC .. format("%02d", c) .. vsR .. "|cffeda55fc|r"
			else
				result = vsC .. format("%d", c) .. vsR .. "|cffeda55fc|r"
			end
		end
		if addon.db.global.discardCopper then
			if g == 0 and s == 0 then
				result = vsC .. format("%d", c) .. vsR .. "|cffeda55fc|r"
			end
		end

		if g > 0 then
			result = vsC .. format("%02d", s) .. vsR .. "|cffc7c7cfs|r " .. result
		else
			if s > 0 then
				result = vsC .. format("%d", s) .. vsR .. "|cffc7c7cfs|r " .. result
			end
		end
	end
	if addon.db.global.discardSilver then
		if g == 0 and s > 0 then
			result = vsC .. format("%d", s) .. vsR .. "|cffc7c7cfs|r " .. result
		end
		if g == 0 and s == 0 then
			result = vsC .. format("%d", c) .. vsR .. "|cffeda55fc|r"
		end
	end
	if g > 0 then
		local gs = tostring(g)
		gs = gsub(gs, "(%d)(%d%d%d)$", "%1" .. LARGE_NUMBER_SEPERATOR .. "%2")
		result = vsC .. gs .. vsR .. "|cffffd70ag|r " .. result
	end

	return result
end

------------------------------------------------------------------------------
-- Market Values

-- return a list of price sources
function Item:GetPriceSources()
	Item:DebugPrintf("GetPriceSources()")
	local ps = {}
	ps.VendorSell = L["VendorSell"]
	Item:DebugPrintf("  add VendorSell (%s)", tostring(ps.VendorSell))

	-- for TSM4
	local tsm4 = {}
	if TSM_API and TSM_API.GetPriceSourceKeys and TSM_API.GetPriceSourceDescription then
		local status, res = pcall(TSM_API.GetPriceSourceKeys, tsm4)
		if not status then
			Item:DebugPrintf("ERR~Broken TSM4 API: pcall of TSM_API.GetPriceSourceKeys=%s", tostring(res))
		end
	end

	local tsmSources = {}
	tsmSources["DBMarket"] = 1
	tsmSources["DBMinBuyout"] = 1
	tsmSources["DBHistorical"] = 0
	tsmSources["DBRegionMarketAvg"] = 1
	tsmSources["DBRegionHistorical"] = 0
	tsmSources["DBRegionMinBuyoutAvg"] = 0
	tsmSources["DBRegionSaleAvg"] = 0

	for _, k in ipairs(tsm4) do
		if tsmSources[k] then
			local status, res = pcall(TSM_API.GetPriceSourceDescription, strlower(k))
			if status then
				ps["TSM:" .. k] = "TSM: " .. tostring(res) .. " (" .. k .. ")"
				Item:DebugPrintf("  add %s (%s)", "TSM:" .. k, ps["TSM:" .. k])
			else
				Item:DebugPrintf("Broken TSM4 API: pcall of TSM_API.GetPriceSourceDescription=%s", tostring(res))
				break
			end
		end
	end

	-- for Undermine Journal
	local uj = {}
	if TUJMarketInfo then
		uj["globalMean"] = L["Region Market Average (globalMean)"]
		uj["globalMedian"] = L["Region Market Median (globalMedian)"]
		uj["market"] = L["14-Days Market Average (market)"]
		uj["recent"] = L["3-Days Market Average (recent)"]
	end
	for k, name in pairs(uj) do
		ps["UJ:" .. k] = "UJ: " .. name
		Item:DebugPrintf("  add %s (%s)", "UJ:" .. k, ps["UJ:" .. k])
	end

	-- for OribosExchange
	local oe = {}
	if OEMarketInfo then
		oe["region"] = L["Median market price across all realms in this region (region)"]
		oe["market"] = L["Median market price on this AH over the past 4 days (market)"]
	end
	for k, name in pairs(oe) do
		ps["OE:" .. k] = "OE: " .. name
		Item:DebugPrintf("  add %s (%s)", "OE:" .. k, ps["OE:" .. k])
	end

	-- for Auction Database
	local ahdb = {}
	if AuctionDB and AuctionDB.AHGetAuctionInfoByLink then
		ahdb["minBid"] = L["minbid"]
		ahdb["minBuyout"] = L["buyout"]
	end
	for k, name in pairs(ahdb) do
		ps["AHDB:" .. k] = "AHDB: " .. name
		Item:DebugPrintf("  add %s (%s)", "AHDB:" .. k, ps["AHDB:" .. k])
	end

	return ps
end

-- get item value of ItemID/ItemLink/BaseItemString or nil
function Item:GetItemValue(link, source)
	Item:DebugPrintf("GetItemValue(%s, %s)", tostring(link), tostring(source))
	if not link or not source then
		return
	end

	private.itemValueCache = private.itemValueCache or {}
	if private.itemValueCache[tostring(link) .. tostring(source)] then
		Item:DebugPrintf("  = %s (cached)", tostring(private.itemValueCache[tostring(link) .. tostring(source)]))
		return private.itemValueCache[tostring(link) .. tostring(source)]
	end

	-- trigger cache
	Item:GetItemInfo(link)

	if source == "VendorSell" then
		local itemName, itemLink, itemRarity, _, _, _, _, _, _, _, itemSellPrice = Item:GetItemInfo(link)
		if itemName and itemLink and itemRarity then
			Item:DebugPrintf("  = %s", tostring(itemSellPrice))
			if itemSellPrice and itemSellPrice == 0 then
				itemSellPrice = nil
			end
			private.itemValueCache[tostring(link) .. tostring(source)] = itemSellPrice
			return itemSellPrice
		end
		return
	end

	if source:match("^TSM:(.+)$") then
		local itemString
		if TSM_API and TSM_API.ToItemString then
			local status, res = pcall(TSM_API.ToItemString, link)
			if status then
				itemString = res
			else
				Item:DebugPrintf("ERR~Broken TSM4 API: pcall of TSM_API.ToItemString=%s", tostring(res))
			end
		end
		if itemString and TSM_API and TSM_API.GetCustomPriceValue and TSM_API.ToItemString then
			local status, res = pcall(TSM_API.GetCustomPriceValue, strlower(source:match("^TSM:(.+)$")), itemString)
			if status then
				Item:DebugPrintf("  = %s", tostring(res))
				private.itemValueCache[tostring(link) .. tostring(source)] = res
				return res
			else
				Item:DebugPrintf("ERR~Broken TSM4 API: pcall of TSM_API.GetCustomPriceValue=%s", tostring(res))
			end
		end
		return
	end

	if source:match("^UJ:(.+)$") then
		if TUJMarketInfo then
			local s = TUJMarketInfo(link)
			local ujsource = source:match("^UJ:(.+)$")
			if s and type(s[ujsource]) == "number" then
				Item:DebugPrintf("  = %s", tostring(s[ujsource]))
				private.itemValueCache[tostring(link) .. tostring(source)] = s[ujsource]
				return s[ujsource]
			end
		end
		return
	end

	if source:match("^OE:(.+)$") then
		if OEMarketInfo then
			local s = {}
			OEMarketInfo(link,s)
			local oesource = source:match("^OE:(.+)$")
			if s and type(s[oesource]) == "number" then
				Item:DebugPrintf("  = %s", tostring(s[oesource]))
				private.itemValueCache[tostring(link) .. tostring(source)] = s[oesource]
				return s[oesource]
			end
		end
		return
	end

	if source:match("^AHDB:(.+)$") then
		if AuctionDB and AuctionDB.AHGetAuctionInfoByLink then
			local auctionData = AuctionDB:AHGetAuctionInfoByLink(link)
			local ahdbsource = source:match("^AHDB:(.+)$")
			if auctionData and type(auctionData[ahdbsource]) == "number" then
				Item:DebugPrintf("  = %s", tostring(auctionData[ahdbsource]))
				private.itemValueCache[tostring(link) .. tostring(source)] = auctionData[ahdbsource]
				return auctionData[ahdbsource]
			end
		end
		return
	end
end

------------------------------------------------------------------------------
-- Sets self disappearing status text
function Item:SetStatusLine(text, seconds)
	private.statusLine = text
	private.statusSeconds = seconds
end

------------------------------------------------------------------------------
-- GetItemInfo Cache + Delayed Update

private.gotNIL = 0
private.gotNILMaxTry = 5
private.gotNILTry = 5

function Item:ResetNILTry()
	Item:DebugPrintf("ResetNILTry()")
	private.gotNILTry = 5
end

function Item:SecTimer()
	-- Item:DebugPrintf("SecTimer()")

	-- ItemInfoCache
	if private.gotNIL > 0 then
		Item:DebugPrintf("ERR~SecTimer: gotNIL=%s gotNILTry=%s", tostring(private.gotNIL), tostring(private.gotNILTry))
		Item:SetStatusLine(L["Refreshing cache for GetItemInfo()..."], 2)
		private.gotNIL = 0
		if private.gotNILTry > 0 and addon.GUI.display then
			private.gotNILTry = private.gotNILTry - 1
			if addon.GUI.display and addon.GUI.view then
				Item:DebugPrintf("ERR~  -> Loot:FillLogFromBags/DebugLog_UpdateData(%s)", tostring(addon.GUI.view))
				addon.Loot:FillLogFromBags(addon.GUI.view)
				addon.GUI.DebugLog_UpdateData(addon.GUI.view, true)
			end
		end
		-- just to be sure
		if not addon.GUI.display then
			Item:DebugPrintf("ERR~SecTimer: gotNIL reset")
			private.gotNIL = 0
			private.gotNILTry = private.gotNILMaxTry
		end
	end

	-- Statusline
	if private.statusLine and strlen(private.statusLine) > 0 then
		private.statusSeconds = private.statusSeconds or 1
		if private.statusFlag ~= nil and private.statusFlag >= private.statusSeconds then
			private.statusFlag = 0
			private.statusLine = ""
		else
			private.statusFlag = private.statusFlag or 0
			private.statusFlag = private.statusFlag + 1
		end
		addon.GUI:SetStatusLine(private.statusLine)
	end
end

function Item:GetItemInfo(link)
	if not link or type(link) ~= "string" then
		return
	end

	private.itemInfoCache = private.itemInfoCache or {}

	if link:match("Henchant") then
		Item:DebugPrintf("8~GetItemInfo(%s/%s): enchant", tostring(link), tostring(Item:EscapedLink(link)))
		private.itemInfoCache[link] = {}
		return
	end

	if private.itemInfoCache[link] then
		-- Item:DebugPrintf("GetItemInfo(%s): hit", tostring(link))
		return unpack(private.itemInfoCache[link])
	else
		local info = {GetItemInfo(link)}
		-- local cName, cLink, _, cLevel, _, _, _, _, _, _, cVendorSell = unpack(info)
		local cName, _, _, _, _, _, _, _, _, _, _ = unpack(info)

		local eLevel = GetDetailedItemLevelInfo(link)
		if cName and eLevel then
			Item:DebugPrintf("8~GetItemInfo(%s/%s/%s): no hit", tostring(link), tostring(Item:EscapedLink(link)), tostring(eLevel))
			info[4] = eLevel
			private.itemInfoCache[link] = info
			return unpack(info)
		else
			Item:DebugPrintf("ERR~7~GetItemInfo(%s/%s): no data", tostring(link), tostring(Item:EscapedLink(link)))
			private.gotNIL = private.gotNIL + 1
			Item:DebugPrintf("ERR~3~Set gotNIL=%s", tostring(private.gotNIL))
		end
	end
end

function Item:GetItemIsBOP(link)
	local bop
	if link and private.itemInfoCache[link] then
		local name, itemLink, _, _, _, _, _, _, _, _, _, _, _, bindType = unpack(private.itemInfoCache[link])
		if name and itemLink then
			bop = (bindType == LE_ITEM_BIND_ON_ACQUIRE or bindType == LE_ITEM_BIND_QUEST) and 1
		end
	end
	return bop
end

-- EOF
